package be.iminds.ilabt.jfed.experiment.setup.config;

import be.iminds.ilabt.jfed.rspec.adv_based_generator.AdvertisementBasedRspecGeneratorConfig;
import com.fasterxml.jackson.annotation.*;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;
import java.util.Map;

@JsonTypeInfo(
        use = JsonTypeInfo.Id.NAME,
        include = JsonTypeInfo.As.PROPERTY,
        property = "source",
        defaultImpl = GenerateUsingJFedRequestRSpec.class,
        visible = true)
@JsonSubTypes({
        @JsonSubTypes.Type(value = GenerateSimpleRequestRSpec.class, name = "GENERATE_SIMPLE"),
        @JsonSubTypes.Type(value = ProvidedUrlRequestRSpec.class, name = "PROVIDE_URL"),
        @JsonSubTypes.Type(value = ProvidedFileRequestRSpec.class, name = "PROVIDE_FILE"),
        @JsonSubTypes.Type(value = ProvidedContentRequestRSpec.class, name = "PROVIDE_CONTENT"),
        @JsonSubTypes.Type(value = GenerateUsingJFedRequestRSpec.class, name = "GENERATE_USING_JFED"),
        @JsonSubTypes.Type(value = GenerateUsingAdvertisementRequestRSpec.class, name = "GENERATE_USING_ADVERTISEMENT"),
        @JsonSubTypes.Type(value = AlreadyProvisionedRequestRSpec.class, name = "ALREADY_PROVISIONED")
})
@JsonInclude(JsonInclude.Include.NON_DEFAULT)
@JsonIgnoreProperties(value={"@context"})
public abstract class RequestRSpec {
    private static final Logger LOG = LoggerFactory.getLogger(RequestRSpec.class);
    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    public enum RequestRSpecSource { PROVIDE_URL, PROVIDE_FILE, PROVIDE_CONTENT,
        GENERATE_USING_JFED, GENERATE_SIMPLE, GENERATE_USING_ADVERTISEMENT,
        ALREADY_PROVISIONED;

        public boolean isProvide() {
            return this == PROVIDE_URL ||
                    this == PROVIDE_FILE ||
                    this == PROVIDE_CONTENT;
        }
    }

    @Nonnull
    protected final RequestRSpecSource source;

    //used for all methods
    private final boolean autoAssignSpecificNodesUsingListResources;
    private final boolean useNitosLease;

    @JsonCreator
    public RequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                        @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                        @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
        assert source != null;
        this.source = source;
        this.autoAssignSpecificNodesUsingListResources =
                autoAssignSpecificNodesUsingListResources == null ?
                        getDefault().getAutoAssignSpecificNodesUsingListResources() :
                        autoAssignSpecificNodesUsingListResources;
        this.useNitosLease = useNitosLease == null ? getDefault().getUseNitosLease() : useNitosLease;
    }

    public static RequestRSpec getDefault() {
        return new GenerateUsingJFedRequestRSpec();
    }


    @JsonCreator
    public static RequestRSpec createRequestRSpecFromJsonObject(@Nonnull Map<String,Object> props) {
        //TODO: Check if types are correct etc
        RequestRSpecSource source =
                props.get("source") == null ?
                        null :
                        RequestRSpecSource.valueOf((String) props.get("source"));
        Boolean autoAssignSpecificNodesUsingListResources = (Boolean) props.get("autoAssignSpecificNodesUsingListResources");
        Boolean useNitosLease = (Boolean) props.get("useNitosLease");

        LOG.debug("createRequestRSpecFromJsonObject with source="+source);

        if (source == null) {
            LOG.debug("createRequestRSpecFromJsonObject with source=null -> reverting all to defaults");
            return getDefault();
        }

        switch (source) {
            case ALREADY_PROVISIONED: {
                return new AlreadyProvisionedRequestRSpec(
                        source,
                        autoAssignSpecificNodesUsingListResources,
                        useNitosLease);
            }
            case PROVIDE_FILE: {
                return new ProvidedFileRequestRSpec(source,
                        (String) props.get("providedContentSource"),
                        (Boolean) props.get("bindTestedServerToUnboundNodes"),
                        (String) props.get("bindUnboundNodesUrn"),
                        autoAssignSpecificNodesUsingListResources,
                        useNitosLease
                );
            }
            case PROVIDE_URL: {
                return new ProvidedUrlRequestRSpec(source,
                        (String) props.get("providedContentSource"),
                        (Boolean) props.get("bindTestedServerToUnboundNodes"),
                        (String) props.get("bindUnboundNodesUrn"),
                        autoAssignSpecificNodesUsingListResources,
                        useNitosLease
                );
            }
            case PROVIDE_CONTENT: {
                return new ProvidedContentRequestRSpec(source,
                        (String) props.get("providedContentSource"),
                        (Boolean) props.get("bindTestedServerToUnboundNodes"),
                        (String) props.get("bindUnboundNodesUrn"),
                        autoAssignSpecificNodesUsingListResources,
                        useNitosLease
                );
            }
            case GENERATE_USING_ADVERTISEMENT: {
                AdvertisementBasedRspecGeneratorConfig advertisementBasedRspecGeneratorConfig =
                        MAPPER.convertValue(
                                props.get("config"),
                                AdvertisementBasedRspecGeneratorConfig.class);
                return new GenerateUsingAdvertisementRequestRSpec(source,
                        autoAssignSpecificNodesUsingListResources,
                        useNitosLease,
                        advertisementBasedRspecGeneratorConfig);
            }
            case GENERATE_SIMPLE: {
                return new GenerateSimpleRequestRSpec(source,
                        (Integer) props.get("nodeCount"),
                        (String) props.get("nodeSliverType"),
                        (Boolean) props.get("nodeExclusive"),
                        (List<String>) props.get("nodeComponentUrn"),
                        (List<String>) props.get("nodeAnsibleGroup"),
                        autoAssignSpecificNodesUsingListResources,
                        useNitosLease);
            }
            case GENERATE_USING_JFED: {
                return new GenerateUsingJFedRequestRSpec(source,
                        (Integer) props.get("nodeCount"),
                        (String) props.get("jFedResourceClass"),
                        (List<String>) props.get("nodeAnsibleGroup"),
                        autoAssignSpecificNodesUsingListResources,
                        useNitosLease);
            }
            default:
                throw new IllegalArgumentException("Config Error: unsupported rspec.source == "+source);
        }
    }

    @JsonIgnore
    public abstract boolean isValid();

    @JsonProperty
    @Nonnull
    public RequestRSpecSource getSource() {
        return source;
    }

    @JsonProperty
    public boolean getAutoAssignSpecificNodesUsingListResources() {
        return autoAssignSpecificNodesUsingListResources;
    }

    @JsonProperty
    public boolean getUseNitosLease() {
        return useNitosLease;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof RequestRSpec)) return false;

        RequestRSpec that = (RequestRSpec) o;

        if (autoAssignSpecificNodesUsingListResources != that.autoAssignSpecificNodesUsingListResources)
            return false;
        if (useNitosLease != that.useNitosLease) return false;
        if (source != that.source) return false;
        return true;
    }

    @Override
    public int hashCode() {
        int result = source.hashCode();
        result = 31 * result + (autoAssignSpecificNodesUsingListResources ? 1 : 0);
        result = 31 * result + (useNitosLease ? 1 : 0);
        return result;
    }
}
